'use strict';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var OmniTypes = require('./types');
var _ = require('lodash');

/**
 * Each property in an OmniSchema definition will be an instance of OmniField.
 * Only "schema", "name", and "type" is a required parameter for construction.
 * All other properties have defaults.
 */

var OmniField = function () {
	function OmniField(schema, name, type, isArray, label) {
		_classCallCheck(this, OmniField);

		this._schema = schema;
		this._name = name;
		this._type = type;
		this._isArray = isArray;
		this._label = label;
	}

	_createClass(OmniField, [{
		key: 'schema',
		get: function get() {
			return this._schema;
		}
	}, {
		key: 'name',
		get: function get() {
			return this._name;
		}
	}, {
		key: 'type',
		get: function get() {
			return this._type;
		}
	}, {
		key: 'isArray',
		get: function get() {
			if (typeof this._isArray === 'undefined') {
				return false;
			} else {
				return this._isArray;
			}
		}
	}, {
		key: 'isRequired',
		get: function get() {
			var req1 = _.get(this, 'required');
			var req2 = _.get(this, 'validation.required');

			if (req1 || req2) {
				return true;
			} else {
				return false;
			}
		}
	}, {
		key: 'label',
		get: function get() {

			if (!this._label) {
				return _.capitalize(_.startCase(this.name));
			}
			return this._label;
		}
	}, {
		key: 'uiExclude',
		get: function get() {
			return Boolean(_.get(this, 'ui.exclude'));
		}
	}, {
		key: 'minVal',
		get: function get() {
			return _.get(this, 'validation.min');
		}
	}, {
		key: 'maxVal',
		get: function get() {
			return _.get(this, 'validation.max');
		}
	}, {
		key: 'isUnique',
		get: function get() {
			return _.get(this, 'db.unique');
		}
	}, {
		key: 'persistence',
		get: function get() {
			return _.get(this, 'db.persistence');
		}
	}, {
		key: 'foreignKeyField',
		get: function get() {
			return _.get(this, 'db.foreignKeyField');
		}
	}, {
		key: 'defaultValue',
		get: function get() {
			if (this.isArray) {
				return [];
			} else if (typeof this.default !== 'undefined') {
				if (typeof this.default === 'function') {
					return this.default();
				} else {
					return this.default;
				}
			} else {
				return this.type.defaultValue;
			}
		}
	}]);

	return OmniField;
}();

var SchemaCache = {};

var OmniSchemaReference = function (_OmniField) {
	_inherits(OmniSchemaReference, _OmniField);

	function OmniSchemaReference(schema, name, referencedSchemaName, isArray, label) {
		_classCallCheck(this, OmniSchemaReference);

		var _this = _possibleConstructorReturn(this, (OmniSchemaReference.__proto__ || Object.getPrototypeOf(OmniSchemaReference)).call(this, schema, name, null, isArray, label));

		_this._referencedSchemaName = referencedSchemaName;
		return _this;
	}

	_createClass(OmniSchemaReference, [{
		key: 'type',
		get: function get() {
			var schema = SchemaCache[this._referencedSchemaName];
			if (!schema) {
				throw new Error('Unable to locate a compiled schema named ' + this._referencedSchemaName);
			}
			return schema;
		}
	}]);

	return OmniSchemaReference;
}(OmniField);

var OmniSchema = function () {
	function OmniSchema(collectionName, parentSchema) {
		_classCallCheck(this, OmniSchema);

		if (collectionName) {

			if (SchemaCache[collectionName]) {
				throw new Error('An OmniSchema collection named ' + collectionName + ' has already been defined. Collection names must be unqiue.');
			}

			Object.defineProperty(this, '__collectionName', { enumerable: false, value: collectionName, writeable: true, configurable: true });

			SchemaCache[collectionName] = this;
		}

		if (parentSchema) {
			Object.defineProperty(this, '__parentSchema', { enumerable: false, value: parentSchema, writeable: true, configurable: true });
			parentSchema.hasChildren = true;
		}
	}

	_createClass(OmniSchema, [{
		key: 'getField',


		/**
  * Returns the OmniField object for the specified field name, or undefined if
  * the field does not exist. If this schema is a subClass, parent schemas are
  * also checked for the field name if it can not be found in the current schema.
  */
		value: function getField(fieldName) {
			var field = this[fieldName];
			if (!field) {
				if (this.parentSchema) {
					field = this.parentSchema.getField(fieldName);
				}
			}
			return field;
		}

		/**
   * Calls the callBack function for each field definition in the schema, passing
   * the OmniField definition as the sole parameter.  If uiOnly is TRUE,
   * the callback will only be called for fields that are marked as part of the UI
   * (i.e. do NOT have the ui.exclude property set to true).
   */

	}, {
		key: 'forEachField',
		value: function forEachField(callBack, uiOnly, excludeInherited) {
			var fl = this.getFieldList(uiOnly, excludeInherited);
			var _iteratorNormalCompletion = true;
			var _didIteratorError = false;
			var _iteratorError = undefined;

			try {
				for (var _iterator = fl[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
					var fieldName = _step.value;

					var field = this.getField(fieldName);
					if (field instanceof OmniSchema.OmniField && (!uiOnly || !field.uiExclude)) {
						callBack(field);
					}
				} // for
			} catch (err) {
				_didIteratorError = true;
				_iteratorError = err;
			} finally {
				try {
					if (!_iteratorNormalCompletion && _iterator.return) {
						_iterator.return();
					}
				} finally {
					if (_didIteratorError) {
						throw _iteratorError;
					}
				}
			}
		}

		/**
   * Returns an array of field names for the defined fields in this schema.
   * if uiOnly is TRUE, the list returned will be for only those fields
   * that are not excluded from the user interface with the 'ui.exclude'
   * property.
   */

	}, {
		key: 'getFieldList',
		value: function getFieldList(uiOnly, excludeInherited) {

			var propName = void 0;
			if (uiOnly) {
				propName = '__uiKeyList';
			} else {
				propName = '__fullKeyList';
			}

			// Return a memoirized list
			if (!this[propName]) {
				var uiKeys = [];
				if (!excludeInherited && this.isSubClass) {
					uiKeys = uiKeys.concat(this.parentSchema.getFieldList(uiOnly));
				}

				for (var fieldName in this) {
					var field = this[fieldName];
					if (field instanceof OmniSchema.OmniField && (!uiOnly || !field.uiExclude)) {
						uiKeys.push(field.name);
					}
				} // for

				Object.defineProperty(this, propName, { configurable: true,
					enumerable: false,
					writeable: true,
					value: uiKeys });
			}

			return this[propName];
		}

		/**
  * Returns a copy of obj that contains only fields that are present in this
  * schema.  If uiOnly is TRUE, the returned object will include only
  * fields that are part of the UI definition (i.e. do NOT have the ui.exclude
  * property set to TRUE). If obj is not an object, "undefined" is returned.
  */

	}, {
		key: 'sanitize',
		value: function sanitize(obj, uiOnly) {
			var _this2 = this;

			if ((typeof obj === 'undefined' ? 'undefined' : _typeof(obj)) !== 'object') {
				return undefined;
			}

			var fieldList = this.getFieldList(uiOnly);
			var result = {};

			var _iteratorNormalCompletion2 = true;
			var _didIteratorError2 = false;
			var _iteratorError2 = undefined;

			try {
				var _loop = function _loop() {
					var fieldName = _step2.value;

					var fieldDef = _this2.getField(fieldName);
					if (fieldDef instanceof OmniSchema.OmniField && (!uiOnly || !fieldDef.uiExclude)) {
						var fieldVal = obj[fieldName];
						if (fieldDef.type instanceof OmniSchema) {
							if (Array.isArray(fieldVal) && fieldDef.isArray) {
								(function () {
									var array = [];
									fieldVal.forEach(function (element) {
										array.push(fieldDef.type.sanitize(element, uiOnly));
									});
									result[fieldName] = array;
								})();
							} else {
								if (fieldDef.isArray && !Array.isArray(fieldVal)) {
									result[fieldName] = [fieldDef.type.sanitize(fieldVal, uiOnly)];
								} else {
									result[fieldName] = fieldDef.type.sanitize(fieldVal, uiOnly);
								}
							}
						} else {
							if (fieldDef.isArray && !Array.isArray(fieldVal)) {
								result[fieldName] = [fieldVal];
							} else {
								result[fieldName] = fieldVal;
							}
						}
					}
				};

				for (var _iterator2 = fieldList[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
					_loop();
				} // for
			} catch (err) {
				_didIteratorError2 = true;
				_iteratorError2 = err;
			} finally {
				try {
					if (!_iteratorNormalCompletion2 && _iterator2.return) {
						_iterator2.return();
					}
				} finally {
					if (_didIteratorError2) {
						throw _iteratorError2;
					}
				}
			}

			return result;
		}

		/**
   * Examines each field in obj and attempts to convert it to
   * its appropriate type if necessary and possible. The new
   * converted object is returned (obj is left untouched).
   * If there are fields in obj that are not part of the
   * schema, they are left untouched unless the sanitize
   * parameter is true, in which case the field will be left
   * out of the returned object. Type conversion is predicated
   * on the schema data type having a fromString() method available
   * to convert it to its appropriate value.
   */

	}, {
		key: 'convert',
		value: function convert(obj, sanitize) {
			var _this3 = this;

			var model = {};

			// eslint-disable-next-line

			var _loop2 = function _loop2(_fieldName) {
				var field = _this3.getField(_fieldName);
				if (field instanceof OmniSchema.OmniField) {
					var modelVal = obj[_fieldName];
					if (modelVal === null) {
						model[_fieldName] = null;
					} else if (!field.isArray) {
						// Process a single value...
						model[_fieldName] = _this3.convertValue(modelVal, field, sanitize);
					} else {
						// Process an array of values...
						if (Array.isArray(modelVal)) {
							model[_fieldName] = [];
							modelVal.forEach(function (entry) {
								model[_fieldName].push(this.convertValue(entry, field, sanitize));
							}.bind(_this3));
						} else if (!sanitize) {
							model[_fieldName] = [_this3.convertValue(modelVal, field, sanitize)];
						}
					}
				} else if (!sanitize) {
					model[_fieldName] = obj[_fieldName];
				}
			};

			for (var _fieldName in obj) {
				_loop2(_fieldName);
			} // for

			return model;
		}
	}, {
		key: 'convertValue',
		value: function convertValue(modelVal, field, sanitize) {

			if (modelVal !== null && modelVal !== undefined) {
				if (field.type instanceof OmniSchema) {
					return field.type.convert(modelVal, sanitize);
				} else if ((typeof modelVal === 'undefined' ? 'undefined' : _typeof(modelVal)) !== field.type.jsType && typeof field.type.fromString === 'function') {
					return field.type.fromString(modelVal.toString());
				}
			}

			return modelVal;
		}

		/**
   * Creates a record that matches the schema, using the fields default values.
   * if uiOnly is TRUE, the returned record will include only fields that are part
   * of the UI definition.
   */

	}, {
		key: 'getDefaultRecord',
		value: function getDefaultRecord(uiOnly) {
			var rec = {};
			this.forEachField(function (field) {
				rec[field.name] = field.defaultValue;
			}, uiOnly);
			return rec;
		}

		/**
   * DataTypeMatches is a helper function for mixin() that is used
   * to determine if a specific data type matches the "match specification".
   * A match specification can be:
   * 1. An object - each property of the object must be present in the data type and the values must match
   * 2. A string - the specified string is a path to a property that must exist in the data type (indenpendent)
   *      of its value.
   * 3. A function - a function that takes a single parameter (the data type), and returns TRUE if the
   *      data type matches
   * 4. An array consisting of one or more of the above 3 types.  ALL entries must match for the data type
   *      to be considered a match (i.e. joined with "logical And").
   * 5. An object with a single member property named "$or", "$and", or $not. These are expected to have a value
   *      that is an array of two or more of the above entries. The entries will be joined with logical
   *      "or", "and", or "not" (i.e. negated truth)
   */

	}, {
		key: 'collectionName',
		get: function get() {
			return this.__collectionName;
		}
	}, {
		key: 'parentSchema',
		get: function get() {
			return this.__parentSchema;
		}
	}, {
		key: 'isSubClass',
		get: function get() {
			return this.hasOwnProperty('__parentSchema');
		}
	}, {
		key: 'hasChildren',
		get: function get() {
			return this.__hasChildren;
		},
		set: function set(isAParent) {
			if (!this.hasOwnProperty('__hasChildren')) {
				Object.defineProperty(this, '__hasChildren', { enumerable: false, value: isAParent, writeable: true, configurable: true });
			} else {
				this.__hasChildren = isAParent;
			}
		}
	}, {
		key: 'rootSchema',
		get: function get() {
			if (this.parentSchema) {
				return this.parentSchema.rootSchema;
			} else {
				return this;
			}
		}
	}], [{
		key: 'compile',


		/**
   * An OmniSchema factory method that compiles a schema template and returns
   * an OmniSchema definition.
   * @param templateObj {object} The schema descriptor to translate into
   * an OmniSchema instance. 
   * @param collectionName {String} The name of a collection that holds this schema (optional).
   * Primarily for use by persistence plugins
   * @param parentSchema {OmniSchema} For object inheritance if this schema inherits field
   * definitions from another schema (optional)
   * Here is an example template:
   * <code>
   *	{	firstName: { type: 'FirstName', required: true },
   *		lastName: { type: OmniTypes.LastName }, 
   *		phones: [{ type: 'Phone' }],
   *		email: { type: 'Email', required: true },
   *		birthdate: { type: 'Date' },
   *		ownerId: { type: 'Integer', ui: { exclude: true } }
   *	}
   *</code>
   */
		value: function compile(templateObj, collectionName, parentSchema) {

			var schema = new OmniSchema(collectionName, parentSchema);

			// eslint-disable-next-line
			for (var _fieldName2 in templateObj) {

				var prop = templateObj[_fieldName2];
				var isArray = void 0;
				if (_.isArray(prop)) {
					isArray = true;
					prop = prop[0];
				} else {
					isArray = false;
				}

				var omniType = void 0,
				    required = void 0,
				    label = void 0,
				    otherProps = void 0,
				    schemaRef = void 0;

				if (prop instanceof OmniTypes.DataType || prop instanceof OmniSchema) {
					omniType = prop;
					otherProps = {};
				} else {
					var type = prop.type;
					schemaRef = prop.schema;
					label = prop.label;

					if (type instanceof OmniTypes.DataType || type instanceof OmniSchema) {
						omniType = type;
					} else if (schemaRef) {
						if (schemaRef instanceof OmniSchema) {
							omniType = schemaRef;
							schemaRef = null;
						} else if (typeof schemaRef !== 'string') {
							throw new Error('\'schema\' property on field ' + _fieldName2 + ' must be a string or an OmniSchema instance.');
						}
					} else {
						omniType = OmniTypes[type];
						if (!omniType) {
							throw new Error('Could not locate omniType named ' + type);
						}
					}

					otherProps = _.omit(prop, ['type', 'label', 'schema']);
				}

				// Create the field definition...
				var omniField = void 0;
				if (schemaRef) {
					omniField = new OmniSchemaReference(schema, _fieldName2, schemaRef, isArray, label);
				} else {
					omniField = new OmniField(schema, _fieldName2, omniType, isArray, label);
				}

				// Merge in any other properties not already compiled...
				Object.assign(omniField, otherProps);

				schema[_fieldName2] = omniField;
			} // for

			return schema;
		}
	}, {
		key: 'DataTypeMatches',
		value: function DataTypeMatches(dataType, matchSpec) {
			if (typeof matchSpec === 'undefined') {
				return typeof dataType.jsType !== 'undefined';
			} else if (typeof matchSpec === 'function') {
				return matchSpec(dataType);
			} else if (typeof matchSpec === 'string') {
				return _.get(dataType, matchSpec);
			} else if (Array.isArray(matchSpec)) {
				for (var i = 0; i < matchSpec.length; i++) {
					var subSpec = matchSpec[i];
					if (!OmniSchema.DataTypeMatches(dataType, subSpec)) {
						return false;
					}
				} // for
				return true;
			} else if ((typeof matchSpec === 'undefined' ? 'undefined' : _typeof(matchSpec)) === 'object') {

				if (matchSpec.hasOwnProperty('$or') && Array.isArray(matchSpec['$or'])) {
					var orList = matchSpec['$or'];
					for (var _i = 0; _i < orList.length; _i++) {
						var _subSpec = orList[_i];
						if (OmniSchema.DataTypeMatches(dataType, _subSpec)) {
							return true;
						}
					} // for
					return false;
				}

				if (matchSpec.hasOwnProperty('$and') && Array.isArray(matchSpec['$and'])) {
					// Using $and is a long hand for the already specified array processing above
					return OmniSchema.DataTypeMatches(dataType, matchSpec['$and']);
				}

				if (matchSpec.hasOwnProperty('$not')) {
					// Return a logical "not" of the specification.
					return !OmniSchema.DataTypeMatches(dataType, matchSpec['$not']);
				}
			}

			return _.isMatch(dataType, matchSpec);
		}

		/**
   * mixin() is the primary means for plugins to extend the functionality of
   * OmniSchema.  It is used to decorate the various base classes of a schema
   * or its data types with additional getters, setters, and/or methods.  See
   * the readme.me file for specifics on the mixin specification.
   * @param spec {object} The mixin specification that adheres to the OmniSchema
   *   plugin mixin specification
   */

	}, {
		key: 'mixin',
		value: function mixin(spec) {

			// Apply part of a specification to a specific object.
			function applySpec(obj, subSpec) {

				if (_.has(subSpec, 'matches')) {
					// This is a conditional subSpec - see if it matches first...
					if (!OmniSchema.DataTypeMatches(obj, subSpec.matches)) {
						// It does not match, so do not apply the subSpec...
						return;
					}
				}

				if (_.has(subSpec, 'get')) {
					var name = subSpec.get.name;
					Object.defineProperty(obj, name, { get: subSpec.get });
				}

				if (_.has(subSpec, 'set')) {
					var _name = subSpec.set.name;
					Object.defineProperty(obj, _name, { set: subSpec.set });
				}

				if (_.has(subSpec, 'func')) {
					var _name2 = subSpec.func.name;
					obj[_name2] = subSpec.func;
				}
			}

			// Make sure obj is an array so we can iterate over it if there
			// is more than one.
			function castArray(obj) {
				if (_.isArray(obj)) {
					return obj;
				} else {
					return [obj];
				}
			}

			if (_.has(spec, 'onSchema')) {
				castArray(spec.onSchema).forEach(function (subSpec) {
					applySpec(OmniSchema.prototype, subSpec);
				});
			}

			if (_.has(spec, 'onField')) {
				castArray(spec.onField).forEach(function (subSpec) {
					applySpec(OmniField.prototype, subSpec);
				});
			}

			if (_.has(spec, 'onType')) {
				castArray(spec.onType).forEach(function (subSpec) {
					_.forEach(OmniTypes, function (value, key, obj) {
						applySpec(value, subSpec);
					});
				});
			}
		}
	}, {
		key: 'Types',
		get: function get() {
			return OmniTypes;
		}
	}]);

	return OmniSchema;
}();

OmniSchema.OmniField = OmniField;

module.exports = OmniSchema;